from inewave.newave.fpha_eco import FphaEco

from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.eco_fpha import MockEcoFpha

ARQ_TESTE = "./tests/mocks/arquivos/__init__.py"


def test_atributos_encontrados_fpha_eco():
    m: MagicMock = mock_open(read_data="".join(MockEcoFpha))
    with patch("builtins.open", m):
        rel = FphaEco.read(ARQ_TESTE)
        assert rel.versao == "FPHA_NEWAVE"
        assert rel.tabela.at[0, "codigo_usina"] == 4
        assert rel.tabela.at[0, "periodo"] == 1
        assert rel.tabela.at[0, "nome_usina"] == "FUNIL-GRANDE"
        assert rel.tabela.at[0, "tipo"] == 2
        assert rel.tabela.at[0, "conv"] == 1
        assert rel.tabela.at[0, "alfa"] == 1
        assert rel.tabela.at[0, "rems"] == 1
        assert rel.tabela.at[0, "numero_pontos_vazao_turbinada"] == 5
        assert rel.tabela.at[0, "vazao_turbinada_minima"] == 0.0
        assert rel.tabela.at[0, "vazao_turbinada_maxima"] == 507.2
        assert rel.tabela.at[0, "numero_pontos_volume_armazenado"] == 1
        assert rel.tabela.at[0, "volume_armazenado_minimo"] == 265.9
        assert rel.tabela.at[0, "volume_armazenado_maximo"] == 265.9
        assert rel.tabela.at[0, "geracao_minima"] == 0.0
        assert rel.tabela.at[0, "geracao_maxima"] == 180.0


def test_eq_fpha_eco():
    m: MagicMock = mock_open(read_data="".join(MockEcoFpha))
    with patch("builtins.open", m):
        rel1 = FphaEco.read(ARQ_TESTE)
        rel2 = FphaEco.read(ARQ_TESTE)
        assert rel1 == rel2


def test_neq_fpha_eco():
    m: MagicMock = mock_open(read_data="".join(MockEcoFpha))
    with patch("builtins.open", m):
        rel1 = FphaEco.read(ARQ_TESTE)
        rel2 = FphaEco.read(ARQ_TESTE)
        rel1.tabela.iloc[0, 0] = -1
        assert rel1 != rel2



from cfinterface.components.section import Section
from cfinterface.components.line import Line
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.floatfield import FloatField
from cfinterface.components.datetimefield import DatetimeField
from typing import List, IO
from datetime import datetime
import pandas as pd  # type: ignore


class BlocoManutencaoUTE(Section):
    """
    Bloco de informações de manutenção das térmicas
    no arquivo do NEWAVE `manutt.dat`.
    """

    __slots__ = ["__linha_uhe", "__cabecalhos"]

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha_uhe = Line(
            [
                IntegerField(2, 0),
                LiteralField(11, 2),
                IntegerField(3, 17),
                LiteralField(13, 21),
                IntegerField(2, 37),
                DatetimeField(8, 40, format="%d%m%Y"),
                IntegerField(3, 49),
                FloatField(7, 55, 2),
            ]
        )
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoManutencaoUTE):
            return False
        bloco: BlocoManutencaoUTE = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO, *args, **kwargs):
        def transforma_em_tabela() -> pd.DataFrame:
            # Converte as informações de cada linha em colunas
            dados = {
                "codigo_empresa": codigos_empresa,
                "nome_empresa": nomes_empresa,
                "codigo_usina": codigos_usina,
                "nome_usina": nomes_usina,
                "codigo_unidade": codigos_unidade,
                "data_inicio": data_inicio,
                "duracao": duracao,
                "potencia": potencia,
            }
            return pd.DataFrame(data=dados)

        # Salta as linhas adicionais
        for _ in range(2):
            self.__cabecalhos.append(file.readline())

        # Para cada usina, lê e processa as informações
        codigos_empresa: List[int] = []
        nomes_empresa: List[str] = []
        codigos_usina: List[int] = []
        nomes_usina: List[str] = []
        codigos_unidade: List[int] = []
        data_inicio: List[datetime] = []
        duracao: List[int] = []
        potencia: List[float] = []
        while True:
            linha = file.readline()
            # Confere se terminaram as usinas
            if len(linha) < 3:
                if len(codigos_empresa) > 0:
                    # Converte para df e salva na variável
                    self.data = transforma_em_tabela()
                break
            dados_uhe = self.__linha_uhe.read(linha)
            codigos_empresa.append(dados_uhe[0])
            nomes_empresa.append(dados_uhe[1])
            codigos_usina.append(dados_uhe[2])
            nomes_usina.append(dados_uhe[3])
            codigos_unidade.append(dados_uhe[4])
            data_inicio.append(dados_uhe[5])
            duracao.append(dados_uhe[6])
            potencia.append(dados_uhe[7])

    # Override
    def write(self, file: IO, *args, **kwargs):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, pd.DataFrame):
            raise ValueError("Dados do manutt.dat não foram lidos com sucesso")

        for _, lin in self.data.iterrows():
            file.write(self.__linha_uhe.write(lin.tolist()))



from inewave.nwlistop.modelos.blocos.submercado import Submercado
from inewave.nwlistop.modelos.arquivos.arquivoclassetermicasubmercadopatamar import (
    ArquivoClasseTermicaSubmercadoPatamar,
)
from inewave.nwlistop.modelos.gtert import GTAnos


class Gtert(ArquivoClasseTermicaSubmercadoPatamar):
    """
    Armazena os dados das saídas referentes à geração térmica por classe
    térmica, por patamar, por submercado.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `gtert00x.out`, onde x varia conforme o
    Submercado em questão.

    """

    BLOCKS = [
        Submercado,
        GTAnos,
    ]



from inewave.nwlistop.modelos.blocos.usina import Usina
from inewave.nwlistop.modelos.arquivos.arquivousinapatamar import (
    ArquivoUsinaPatamar,
)
from inewave.nwlistop.modelos.ghmax_fpha import GhmaxfphaAnos


class GhmaxFphc(ArquivoUsinaPatamar):
    """
    Armazena os dados das saídas referentes à geração hidráulica máxima
    no ponto de acesso à FPH constante da usina.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `ghmax_fphcXXX.out`, onde x varia conforme
    a usina em questão.

    """

    BLOCKS = [
        Usina,
        GhmaxfphaAnos,
    ]



from inewave.nwlistop.modelos.blocos.restricao import Restricao
from inewave.nwlistop.modelos.arquivos.arquivorestricaopatamar import (
    ArquivoRestricaoPatamar,
)
from inewave.nwlistop.modelos.c_v_rhq import CVRHQAnos

from warnings import warn


class CVRHQ(ArquivoRestricaoPatamar):
    """
    Armazena os dados das saídas referentes ao custo de violação
    de RHQ por restrição.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `c_v_rhqXXX.out`.

    """

    BLOCKS = [
        Restricao,
        CVRHQAnos,
    ]

    def __init__(self, data=...) -> None:
        warn(
            "Esta classe é relativa a um arquivo que não é mais suportado."
            + " Utilize a classe CviolRhq no lugar.",
            DeprecationWarning,
        )
        super().__init__(data)



from cfinterface.components.block import Block
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.line import Line

from typing import IO, List


class VersaoModelo(Block):
    """
    Bloco para ler a versão utilizada do modelo a partir da linha de
    título do arquivo.
    """

    __slots__: List[str] = []

    BEGIN_PATTERN = r"CEPEL: NEWAVE"
    END_PATTERN = ""

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, VersaoModelo):
            return False
        else:
            if not all([type(self.data) is str, type(o.data) is str]):
                return False
            return self.data == o.data

    def read(self, file: IO, *args, **kwargs):
        linha = file.readline()
        modelo_linha = Line([LiteralField(size=12, starting_position=29)])
        self.data = modelo_linha.read(linha)[0]


class VersaoModeloLibs(Block):
    """
    Bloco para ler a versão utilizada do modelo a partir da linha de
    título do arquivo.
    """

    __slots__: List[str] = []

    BEGIN_PATTERN = r"NEWAVE -"
    END_PATTERN = ""

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, VersaoModeloLibs):
            return False
        else:
            if not all([type(self.data) is str, type(o.data) is str]):
                return False
            return self.data == o.data

    def read(self, file: IO, *args, **kwargs):
        linha = file.readline()
        modelo_linha = Line([LiteralField(size=20, starting_position=12)])
        self.data = modelo_linha.read(linha)[0]



from inewave.newave.fpha_avl_desv_s import FphaAvlDesvS

from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.avl_desvfpha_s import MockAvlDesvFphaS

ARQ_TESTE = "./tests/mocks/arquivos/__init__.py"


def test_atributos_encontrados_fpha_avl_desv_s():
    m: MagicMock = mock_open(read_data="".join(MockAvlDesvFphaS))
    with patch("builtins.open", m):
        rel = FphaAvlDesvS.read(ARQ_TESTE)
        assert rel.versao == "FPHA_NEWAVE"
        assert rel.tabela.at[0, "codigo_usina"] == 4
        assert rel.tabela.at[0, "nome_usina"] == "FUNIL-GRANDE"
        assert rel.tabela.at[0, "volume_armazenado_percentual"] == 100.0
        assert rel.tabela.at[0, "vazao_turbinada_m3s"] == 0.0
        assert rel.tabela.at[0, "vazao_vertida_m3s"] == 0.0
        assert rel.tabela.at[0, "desvio_percentual"] == 0.0


def test_es_fpha_avl_desv_s():
    m: MagicMock = mock_open(read_data="".join(MockAvlDesvFphaS))
    with patch("builtins.open", m):
        rel1 = FphaAvlDesvS.read(ARQ_TESTE)
        rel2 = FphaAvlDesvS.read(ARQ_TESTE)
        assert rel1 == rel2


## Não precisa comparar isto, é somente leitura - passa por agregação
# def test_nes_fpha_avl_desv_s():
#     m: MagicMock = mock_open(read_data="".join(MockAvlDesvFphaS))
#     with patch("builtins.open", m):
#         rel1 = AvlDesvFphaS.read(ARQ_TESTE)
#         rel2 = AvlDesvFphaS.read(ARQ_TESTE)
#         rel1.tabela.iloc[0, 0] = -1
#         assert rel1 != rel2



from inewave.nwlistop.modelos.arquivos.arquivosinpatamar import (
    ArquivoSINPatamar,
)
from inewave.nwlistop.modelos.viol_lpp_dfmaxsin import ViolLppDfmaxsinAnos


class ViolLppDfmaxsin(ArquivoSINPatamar):
    """
    Armazena os dados das saídas referentes à violação das restrições
    LPP de defluência máxima por patamar para o SIN.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `viol_lpp_dfmaxsin.out`.
    """

    BLOCKS = [
        ViolLppDfmaxsinAnos,
    ]



from inewave.nwlistop.modelos.arquivos.arquivosinpatamar import (
    ArquivoSINPatamar,
)
from inewave.nwlistop.modelos.viol_ghminsin import ViolGhminsinAnos


class ViolGhminsin(ArquivoSINPatamar):
    """
    Armazena os dados das saídas referentes à violação da meta de geração
    hidraulica mínima por patamar, para o SIN.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas no `viol_ghminsin.out`.

    """

    BLOCKS = [
        ViolGhminsinAnos,
    ]



from inewave.config import MESES_DF

from cfinterface.components.line import Line
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.floatfield import FloatField
from inewave.nwlistop.modelos.blocos.valoresserie import ValoresSerie


class CdefAnos(ValoresSerie):
    """
    Bloco com as informações das tabelas de custos de déficit.
    """

    __slots__ = []

    HEADER_LINE = Line([IntegerField(4, 10)])
    DATA_LINE = Line(
        [  # type: ignore
            IntegerField(4, 2),
        ]
        + [FloatField(10, 7 + 10 * i, 2) for i in range(len(MESES_DF))]  # type: ignore
    )



from inewave.newave.modelos.adterm import BlocoUTEAdTerm
from inewave.newave import Adterm

from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.adterm import MockBlocoUTEAdTerm


ARQ_TESTE = "./tests/mocks/arquivos/__init__.py"


def test_bloco_ute_adterm():
    m: MagicMock = mock_open(read_data="".join(MockBlocoUTEAdTerm))
    b = BlocoUTEAdTerm()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.shape[0] == 18
    assert b.data.shape[1] == 5
    assert b.data.iloc[0, 4] == 230.70
    assert b.data.iloc[-1, -1] == 0.00


def test_atributos_encontrados_adterm():
    m: MagicMock = mock_open(read_data="".join(MockBlocoUTEAdTerm))
    with patch("builtins.open", m):
        ad = Adterm.read(ARQ_TESTE)
        assert ad.despachos is not None


def test_atributos_nao_encontrados_adterm():
    m: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m):
        ad = Adterm.read(ARQ_TESTE)
        assert ad.despachos is None


def test_eq_adterm():
    m: MagicMock = mock_open(read_data="".join(MockBlocoUTEAdTerm))
    with patch("builtins.open", m):
        ad1 = Adterm.read(ARQ_TESTE)
        ad2 = Adterm.read(ARQ_TESTE)
        assert ad1 == ad2


def test_neq_adterm():
    m: MagicMock = mock_open(read_data="".join(MockBlocoUTEAdTerm))
    with patch("builtins.open", m):
        ad1 = Adterm.read(ARQ_TESTE)
        ad2 = Adterm.read(ARQ_TESTE)
        ad2.despachos.iloc[0, 0] = -1
        assert ad1 != ad2


def test_leitura_escrita_adterm():
    m_leitura: MagicMock = mock_open(read_data="".join(MockBlocoUTEAdTerm))
    with patch("builtins.open", m_leitura):
        ad1 = Adterm.read(ARQ_TESTE)
    m_escrita: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m_escrita):
        ad1.write(ARQ_TESTE)
        # Recupera o que foi escrito
        chamadas = m_escrita.mock_calls
        linhas_escritas = [
            chamadas[i].args[0] for i in range(1, len(chamadas) - 1)
        ]
    m_releitura: MagicMock = mock_open(read_data="".join(linhas_escritas))
    with patch("builtins.open", m_releitura):
        ad2 = Adterm.read(ARQ_TESTE)
        assert ad1 == ad2



from inewave.nwlistop.modelos.arquivos.arquivosinpatamar import (  # noqa
    ArquivoSINPatamar,
)
from inewave.nwlistop.modelos.deletricas import DeletricasAnos

from warnings import warn


class Deletricas(ArquivoSINPatamar):
    """
    Armazena os dados das saídas referentes à violação das
    restrições elétricas para o SIN.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas no `deletricas.out`.

    """

    BLOCKS = [
        DeletricasAnos,
    ]

    def __init__(self, data=...) -> None:
        warn(
            "Esta classe é relativa a um arquivo que não é mais suportado."
            + " Utilize a classe ViolEletricas no lugar.",
            DeprecationWarning,
        )
        super().__init__(data)



from inewave.newave.modelos.expt import BlocoUTEExpt

from cfinterface.files.sectionfile import SectionFile
from cfinterface.components.section import Section
from typing import TypeVar, List, Type, Optional
import pandas as pd  # type: ignore


class Expt(SectionFile):
    """
    Armazena os dados de entrada do NEWAVE referentes à expansão
    térmica do sistema.

    """

    T = TypeVar("T")

    SECTIONS: List[Type[Section]] = [BlocoUTEExpt]

    @property
    def expansoes(self) -> Optional[pd.DataFrame]:
        """
        A tabela de expansões das UTEs.

        - codigo_usina (`int`)
        - tipo (`str`)
        - modificacao (`float`)
        - data_inicio (`datetime`)
        - data_fim (`datetime`)
        - nome_usina (`str`)

        :return: A tabela como um DataFrame
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_sections_of_type(BlocoUTEExpt)
        if isinstance(b, BlocoUTEExpt):
            return b.data
        return None

    @expansoes.setter
    def expansoes(self, d: pd.DataFrame):
        b = self.data.get_sections_of_type(BlocoUTEExpt)
        if isinstance(b, BlocoUTEExpt):
            b.data = d



from cfinterface.files.sectionfile import SectionFile
from cfinterface.components.section import Section
from typing import TypeVar, List, Type, Optional

from inewave.newave.modelos.ghmin import BlocoUHEGhmin

import pandas as pd  # type: ignore


class Ghmin(SectionFile):
    """
    Armazena os dados de entrada do NEWAVE referentes à geração hidráulica
    mínima por usina.

    """

    T = TypeVar("T")

    SECTIONS: List[Type[Section]] = [BlocoUHEGhmin]

    @property
    def geracoes(self) -> Optional[pd.DataFrame]:
        """
        Tabela com as gerações mínimas das usinas
        hidráulicas.

        - codigo_usina (`int`)
        - data (`datetime`)
        - patamar (`int`)
        - geracao (`float`)

        :return: A tabela como um DataFrame
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_sections_of_type(BlocoUHEGhmin)
        if isinstance(b, BlocoUHEGhmin):
            return b.data
        return None

    @geracoes.setter
    def geracoes(self, valor: pd.DataFrame):
        b = self.data.get_sections_of_type(BlocoUHEGhmin)
        if isinstance(b, BlocoUHEGhmin):
            b.data = valor
        else:
            raise ValueError("Campo não lido")



# Rotinas de testes associadas ao arquivo selcor.dat do NEWAVE
from inewave.newave.modelos.selcor import BlocoDadosSelcor

from inewave.newave import Selcor


from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.selcor import MockSelcor

ARQ_TESTE = "./tests/mocks/arquivos/__init__.py"


def test_bloco_usinas_selcor():
    m: MagicMock = mock_open(read_data="".join(MockSelcor))
    b = BlocoDadosSelcor()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)
    assert len(b.data) == 7
    assert b.data[0][0] == 2
    assert b.data[1][0] == 0
    assert b.data[2][0] == 12
    assert b.data[3][0] == 1
    assert b.data[4][0] == 0
    assert b.data[5] == [7, 15]
    assert b.data[6] == [2, 9]


def test_atributos_encontrados_selcor():
    m: MagicMock = mock_open(read_data="".join(MockSelcor))
    with patch("builtins.open", m):
        ad = Selcor.read(ARQ_TESTE)
        assert ad.iteracao_inicial is not None
        assert ad.tamanho_janela is not None
        assert ad.numero_cortes_adicionados_por_iteracao is not None
        assert ad.considera_cortes_da_propria_iteracao is not None
        assert ad.imprime_relatorio is not None
        assert ad.iteracoes_impressao is not None
        assert ad.series_impressao is not None


def test_atributos_nao_encontrados_selcor():
    m: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m):
        ad = Selcor.read(ARQ_TESTE)
        assert ad.iteracao_inicial is None
        assert ad.tamanho_janela is None
        assert ad.numero_cortes_adicionados_por_iteracao is None
        assert ad.considera_cortes_da_propria_iteracao is None
        assert ad.imprime_relatorio is None
        assert ad.iteracoes_impressao == [None, None]
        assert ad.series_impressao == [None, None]


def test_eq_selcor():
    m: MagicMock = mock_open(read_data="".join(MockSelcor))
    with patch("builtins.open", m):
        cf1 = Selcor.read(ARQ_TESTE)
        cf2 = Selcor.read(ARQ_TESTE)
        assert cf1 == cf2


def test_neq_selcor():
    m: MagicMock = mock_open(read_data="".join(MockSelcor))
    with patch("builtins.open", m):
        cf1 = Selcor.read(ARQ_TESTE)
        cf2 = Selcor.read(ARQ_TESTE)
        cf2.iteracao_inicial = 1
        assert cf1 != cf2


def test_leitura_escrita_selcor():
    m_leitura: MagicMock = mock_open(read_data="".join(MockSelcor))
    with patch("builtins.open", m_leitura):
        cf1 = Selcor.read(ARQ_TESTE)
    m_escrita: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m_escrita):
        cf1.write(ARQ_TESTE)
        # Recupera o que foi escrito
        chamadas = m_escrita.mock_calls
        linhas_escritas = [
            chamadas[i].args[0] for i in range(1, len(chamadas) - 1)
        ]
    m_releitura: MagicMock = mock_open(read_data="".join(linhas_escritas))
    with patch("builtins.open", m_releitura):
        cf2 = Selcor.read(ARQ_TESTE)
        assert cf1 == cf2



from inewave.nwlistop.modelos.verturbsin import VertAnos
from inewave.nwlistop.modelos.arquivos.arquivosin import (
    ArquivoSIN,
)


class Verturbsin(ArquivoSIN):
    """
    Armazena os dados das saídas referentes às energias
    vertidas para o SIN.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `verturbsin.out`.

    """

    BLOCKS = [
        VertAnos,
    ]



# Rotinas de testes associadas ao arquivo MEDIAS-REE.CSV do NWLISTOP
from inewave.nwlistop.mediasree import Mediasree


ARQ_TESTE = "tests/_arquivos/MEDIAS-REE.CSV"


def test_eq_mediasree():
    leitor = Mediasree.read(ARQ_TESTE)
    leitor2 = Mediasree.read(ARQ_TESTE)
    assert leitor == leitor2
    assert leitor.valores.shape == (720, 49)



from inewave.newave.modelos.enavazb import SecaoDadosEnavazb
from inewave.newave.enavazb import Enavazb


from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch


ARQ_TESTE = "./tests/mocks/arquivos/enavazb.dat"

NUM_FORWARDS = 2
NUM_REES = 1
NUM_ABERTURAS = 20
NUM_ESTAGIOS = 16
NUM_ENTRADAS = NUM_FORWARDS * NUM_ABERTURAS * NUM_REES * NUM_ESTAGIOS


def test_secao_enavaz():
    r = SecaoDadosEnavazb()
    with open(ARQ_TESTE, "rb") as fp:
        r.read(
            fp,
            numero_forwards=NUM_FORWARDS,
            numero_aberturas=NUM_ABERTURAS,
            numero_rees=NUM_REES,
            numero_estagios=NUM_ESTAGIOS,
        )

    assert len(r.data) == NUM_ENTRADAS


def test_atributos_encontrados_enavaz():
    h = Enavazb.read(
        ARQ_TESTE,
        numero_forwards=NUM_FORWARDS,
        numero_aberturas=NUM_ABERTURAS,
        numero_rees=NUM_REES,
        numero_estagios=NUM_ESTAGIOS,
    )
    assert h.series is not None
    assert h.series.isna().sum().sum() == 0


def test_atributos_nao_encontrados_enavaz():
    m: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m):
        h = Enavazb.read(
            ARQ_TESTE,
            numero_forwards=NUM_FORWARDS,
            numero_aberturas=NUM_ABERTURAS,
            numero_rees=NUM_REES,
            numero_estagios=NUM_ESTAGIOS,
        )
        assert h.series.isna().sum().sum() == NUM_ENTRADAS


def test_eq_enavaz():
    h1 = Enavazb.read(
        ARQ_TESTE,
        numero_forwards=NUM_FORWARDS,
        numero_aberturas=NUM_ABERTURAS,
        numero_rees=NUM_REES,
        numero_estagios=NUM_ESTAGIOS,
    )
    h2 = Enavazb.read(
        ARQ_TESTE,
        numero_forwards=NUM_FORWARDS,
        numero_aberturas=NUM_ABERTURAS,
        numero_rees=NUM_REES,
        numero_estagios=NUM_ESTAGIOS,
    )
    assert h1 == h2



from inewave.nwlistop.modelos.arquivos.arquivosin import (
    ArquivoSIN,
)
from inewave.nwlistop.modelos.c_v_rhv_s import CVRHVAnos

from warnings import warn


class CVRHVs(ArquivoSIN):
    """
    Armazena os dados das saídas referentes ao custo de violação
    de RHV para o SIN

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas no `c_v_rhv_s.out`.

    """

    BLOCKS = [
        CVRHVAnos,
    ]

    def __init__(self, data=...) -> None:
        warn(
            "Esta classe é relativa a um arquivo que não é mais suportado."
            + " Utilize a classe CviolRhvsin no lugar.",
            DeprecationWarning,
        )
        super().__init__(data)



# Inclui os membros

from .caso import Caso  # noqa
from .arquivos import Arquivos  # noqa
from .estados import Estados  # noqa
from .nwlistcfdat import Nwlistcfdat  # noqa
from .nwlistcfrel import Nwlistcfrel  # noqa



# Rotinas de testes associadas ao arquivo MEDIAS-RHQ.CSV do NWLISTOP
from inewave.nwlistop.mediasrhq import Mediasrhq


ARQ_TESTE = "tests/_arquivos/MEDIAS-RHQ.CSV"


def test_eq_mediasrhq():
    leitor = Mediasrhq.read(ARQ_TESTE)
    leitor2 = Mediasrhq.read(ARQ_TESTE)
    assert leitor == leitor2
    assert leitor.valores.shape == (0, 2)



from inewave.nwlistop.modelos.blocos.restricao import Restricao
from inewave.nwlistop.modelos.arquivos.arquivorestricao import (
    ArquivoRestricao,
)
from inewave.nwlistop.modelos.viol_rhv import ViolRHVAnos


class ViolRHV(ArquivoRestricao):
    """
    Armazena os dados das saídas referentes aos valores das violações
    das restrições RHV por restrição.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `viol_rhvXXX.out`.

    """

    BLOCKS = [
        Restricao,
        ViolRHVAnos,
    ]



from inewave.nwlistop.modelos.blocos.estacaobombeamento import (
    EstacaoBombeamento,
)
from inewave.nwlistop.modelos.arquivos.arquivoestacaobombeamentopatamar import (  # noqa
    ArquivoEstacaoBombeamentoPatamar,
)
from inewave.nwlistop.modelos.cbomb import CBombAnos


class Cbomb(ArquivoEstacaoBombeamentoPatamar):
    """
    Armazena os dados das saídas referentes ao consumo de energia para
    bombeamento por estação.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `cbombXXX.out`.

    """

    BLOCKS = [
        EstacaoBombeamento,
        CBombAnos,
    ]



# Rotinas de testes associadas ao arquivo modif.dat do NEWAVE
from inewave.newave.modelos.modif import (
    USINA,
    VOLMIN,
    VOLMAX,
    NUMCNJ,
    NUMMAQ,
    VAZMIN,
    CFUGA,
    CMONT,
    VMAXT,
    VMINT,
    VMINP,
    VAZMINT,
    VAZMAXT,
    TURBMAXT,
    TURBMINT,
)

from inewave.newave import Modif

from datetime import datetime
from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.modif import (
    MockModif,
    MockUSINA,
    MockVOLMIN,
    MockVOLMAX,
    MockNUMCNJ,
    MockNUMMAQ,
    MockVAZMIN,
    MockCFUGA,
    MockCMONT,
    MockVMAXT,
    MockVMINT,
    MockVMINP,
    MockVAZMINT,
    MockVAZMAXT,
    MockTURBMAXT,
    MockTURBMINT,
)

ARQ_TESTE = "./tests/mocks/arquivos/__init__.py"


def test_registro_usina_modif():
    m: MagicMock = mock_open(read_data="".join(MockUSINA))
    r = USINA()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [1, "CAMARGOS"]
    assert r.codigo == 1
    r.codigo = 2
    assert r.codigo == 2
    assert r.nome == "CAMARGOS"
    r.nome = "C1"
    assert r.nome == "C1"


def test_registro_vazmin_modif():
    m: MagicMock = mock_open(read_data="".join(MockVAZMIN))
    r = VAZMIN()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [34]
    assert r.vazao == 34
    r.vazao = 50
    assert r.vazao == 50


def test_registro_vmaxt_modif():
    m: MagicMock = mock_open(read_data="".join(MockVMAXT))
    r = VMAXT()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [12, 2021, 61.310, "'%'"]
    assert r.data_inicio == datetime(2021, 12, 1)
    r.data_inicio = datetime(2022, 12, 1)
    assert r.data_inicio == datetime(2022, 12, 1)
    assert r.volume == 61.310
    r.volume = 50.0
    assert r.volume == 50.0
    assert r.unidade == "'%'"
    r.unidade = "'h'"
    assert r.unidade == "'h'"


def test_registro_vazmint_modif():
    m: MagicMock = mock_open(read_data="".join(MockVAZMINT))
    r = VAZMINT()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [10, 2021, 10.00]
    assert r.data_inicio == datetime(2021, 10, 1)
    r.data_inicio = datetime(2022, 10, 1)
    assert r.data_inicio == datetime(2022, 10, 1)
    assert r.vazao == 10.00
    r.vazao = 5.0
    assert r.vazao == 5.0


def test_registro_volmin_modif():
    m: MagicMock = mock_open(read_data="".join(MockVOLMIN))
    r = VOLMIN()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [15563.63, "'h'"]
    assert r.volume == 15563.63
    r.volume = 15
    assert r.volume == 15
    assert r.unidade == "'h'"
    r.unidade = "'%'"
    assert r.unidade == "'%'"


def test_registro_volmax_modif():
    m: MagicMock = mock_open(read_data="".join(MockVOLMAX))
    r = VOLMAX()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [55.0, "'%'"]
    assert r.volume == 55.0
    r.volume = 1000
    assert r.volume == 1000
    assert r.unidade == "'%'"
    r.unidade = "'h'"
    assert r.unidade == "'h'"


def test_registro_numcnj_modif():
    m: MagicMock = mock_open(read_data="".join(MockNUMCNJ))
    r = NUMCNJ()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [2]
    assert r.numero == 2
    r.numero = 1
    assert r.numero == 1


def test_registro_nummaq_modif():
    m: MagicMock = mock_open(read_data="".join(MockNUMMAQ))
    r = NUMMAQ()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [2, 5]
    assert r.conjunto == 2
    r.conjunto = 1
    assert r.conjunto == 1
    assert r.numero_maquinas == 5
    r.numero_maquinas = 10
    assert r.numero_maquinas == 10


def test_registro_vmint_modif():
    m: MagicMock = mock_open(read_data="".join(MockVMINT))
    r = VMINT()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [10, 2021, 20.0, "'%'"]
    assert r.data_inicio == datetime(2021, 10, 1)
    r.data_inicio = datetime(2022, 10, 1)
    assert r.data_inicio == datetime(2022, 10, 1)
    assert r.volume == 20.0
    r.volume = 10.0
    assert r.volume == 10.0
    assert r.unidade == "'%'"
    r.unidade = "'h'"
    assert r.unidade == "'h'"


def test_registro_vminp_modif():
    m: MagicMock = mock_open(read_data="".join(MockVMINP))
    r = VMINP()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [10, 2021, 20.0, "'%'"]
    assert r.data_inicio == datetime(2021, 10, 1)
    r.data_inicio = datetime(2022, 10, 1)
    assert r.data_inicio == datetime(2022, 10, 1)
    assert r.volume == 20.0
    r.volume = 10.0
    assert r.volume == 10.0
    assert r.unidade == "'%'"
    r.unidade = "'h'"
    assert r.unidade == "'h'"


def test_registro_cfuga_modif():
    m: MagicMock = mock_open(read_data="".join(MockCFUGA))
    r = CFUGA()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [10, 2021, 5.60]
    assert r.data_inicio == datetime(2021, 10, 1)
    r.data_inicio = datetime(2022, 10, 1)
    assert r.data_inicio == datetime(2022, 10, 1)
    assert r.nivel == 5.60
    r.nivel = 10.0
    assert r.nivel == 10.0


def test_registro_cmont_modif():
    m: MagicMock = mock_open(read_data="".join(MockCMONT))
    r = CMONT()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [10, 2021, 71.30]
    assert r.data_inicio == datetime(2021, 10, 1)
    r.data_inicio = datetime(2022, 10, 1)
    assert r.data_inicio == datetime(2022, 10, 1)
    assert r.nivel == 71.30
    r.nivel = 10.0
    assert r.nivel == 10.0


def test_registro_vazmaxt_modif():
    m: MagicMock = mock_open(read_data="".join(MockVAZMAXT))
    r = VAZMAXT()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [1, 2020, 0.00]
    assert r.data_inicio == datetime(2020, 1, 1)
    r.data_inicio = datetime(2022, 10, 1)
    assert r.data_inicio == datetime(2022, 10, 1)
    assert r.vazao == 0.0
    r.vazao = 5.0
    assert r.vazao == 5.0


def test_registro_turbmaxt_modif():
    m: MagicMock = mock_open(read_data="".join(MockTURBMAXT))
    r = TURBMAXT()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [1, 2020, 0.00]
    assert r.data_inicio == datetime(2020, 1, 1)
    r.data_inicio = datetime(2022, 10, 1)
    assert r.data_inicio == datetime(2022, 10, 1)
    assert r.turbinamento == 0.0
    r.turbinamento = 5.0
    assert r.turbinamento == 5.0


def test_registro_turbmint_modif():
    m: MagicMock = mock_open(read_data="".join(MockTURBMINT))
    r = TURBMINT()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [1, 2020, 0.00]
    assert r.data_inicio == datetime(2020, 1, 1)
    r.data_inicio = datetime(2022, 10, 1)
    assert r.data_inicio == datetime(2022, 10, 1)
    assert r.turbinamento == 0.0
    r.turbinamento = 5.0
    assert r.turbinamento == 5.0


def test_atributos_encontrados_modif():
    m: MagicMock = mock_open(read_data="".join(MockModif))
    with patch("builtins.open", m):
        ad = Modif.read(ARQ_TESTE)
        assert len(ad.usina()) > 0
        assert len(ad.vazmin()) > 0
        assert len(ad.vmaxt()) > 0
        assert len(ad.vazmint()) > 0
        assert isinstance(ad.volmin(), VOLMIN)
        assert isinstance(ad.volmax(), VOLMAX)
        assert len(ad.numcnj()) > 0
        assert len(ad.nummaq()) > 0
        assert len(ad.vmint()) > 0
        assert len(ad.vminp()) > 0
        assert len(ad.cfuga()) > 0
        assert len(ad.cmont()) > 0
        assert len(ad.vazmaxt()) > 0
        assert len(ad.turbmaxt()) > 0
        assert len(ad.turbmint()) > 0


def test_leitura_escrita_modif():
    m_leitura: MagicMock = mock_open(read_data="".join(MockModif))
    with patch("builtins.open", m_leitura):
        cf1 = Modif.read(ARQ_TESTE)
    m_escrita: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m_escrita):
        cf1.write(ARQ_TESTE)
        # Recupera o que foi escrito
        chamadas = m_escrita.mock_calls
        linhas_escritas = [
            chamadas[i].args[0] for i in range(1, len(chamadas) - 1)
        ]

    m_releitura: MagicMock = mock_open(read_data="".join(linhas_escritas))
    with patch("builtins.open", m_releitura):
        cf2 = Modif.read(ARQ_TESTE)
        assert cf1 == cf2



from inewave.nwlistop.modelos.blocos.usina import Usina
from inewave.nwlistop.modelos.arquivos.arquivousinapatamar import (
    ArquivoUsinaPatamar,
)
from inewave.nwlistop.modelos.dvazmax import DvazmaxAnos

from warnings import warn


class Dvazmax(ArquivoUsinaPatamar):
    """
    Armazena os dados das saídas referentes à violação de restrição de
    vazão máxima por usina.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `dvazmax00x.out`, onde x varia conforme
    a usina em questão.

    """

    BLOCKS = [
        Usina,
        DvazmaxAnos,
    ]

    def __init__(self, data=...) -> None:
        warn(
            "Esta classe é relativa a um arquivo que não é mais suportado."
            + " Utilize a classe ViolVazmax no lugar.",
            DeprecationWarning,
        )
        super().__init__(data)



from cfinterface.files.sectionfile import SectionFile
from inewave.newave.modelos.cortes import SecaoDadosCortes

import pandas as pd  # type: ignore
from typing import TypeVar, Optional, Union, List


class Cortes(SectionFile):
    """
    Armazena os dados de saída do NEWAVE referentes aos
    cortes de Benders.
    """

    T = TypeVar("T")

    SECTIONS = [SecaoDadosCortes]
    STORAGE = "BINARY"

    @classmethod
    def read(
        cls,
        content: Union[str, bytes],
        tamanho_registro: int = 1664,
        indice_ultimo_corte: int = 0,
        numero_total_cortes: int = 10000,
        codigos_rees: List[int] = [],
        codigos_uhes: List[int] = [],
        codigos_submercados: List[int] = [],
        ordem_maxima_parp: int = 12,
        lag_maximo_gnl: int = 2,
        *args,
        **kwargs
    ) -> "Cortes":
        return super().read(
            content,
            tamanho_registro=tamanho_registro,
            indice_ultimo_corte=indice_ultimo_corte,
            numero_total_cortes=numero_total_cortes,
            codigos_rees=codigos_rees,
            codigos_uhes=codigos_uhes,
            codigos_submercados=codigos_submercados,
            ordem_maxima_parp=ordem_maxima_parp,
            lag_maximo_gnl=lag_maximo_gnl,
            *args,
            **kwargs
        )

    @property
    def cortes(self) -> Optional[pd.DataFrame]:
        """
        Retorna o conjunto dos cortes de Benders construídos
        durante o cálculo da política.

        - indice_corte (`int`)
        - iteracao_construcao (`str`)
        - indice_forward (`int`)
        - iteracao_inativacao (`int`)
        - rhs (`float`)

        Se o estágio em questão for agregado em REE:

        - pi_earm_ree1 (`float`)
        - pi_ena_ree1_lag1 (`float`)
        - ...
        - pi_ena_ree1_lagN (`float`)
        - pi_earm_ree2 (`float`)
        - ...
        - pi_ena_reeR_lagN (`float`)

        Se o estágio em questão for individualizado:

        - pi_varm_uhe1 (`float`)
        - pi_qafl_uhe1_lag1 (`float`)
        - ...
        - pi_qafl_uhe1_lagN (`float`)
        - pi_varm_uhe2 (`float`)
        - ...
        - pi_qafl_uheU_lagN (`float`)

        Para todos os estágios:

        - pi_gnl_sbm1_pat1_lag1 (`float`)
        - ...
        - pi_gnl_sbm1_pat1_lagL (`float`)
        - pi_gnl_sbm1_pat2_lag1 (`float`)
        - ...
        - pi_gnl_sbm1_patP_lagL (`float`)
        - pi_gnl_sbm2_pat1_lag1 (`float`)
        - ...
        - pi_gnl_sbmS_patP_lagL (`float`)

        R é o número de REEs.

        U é o número de UHEs.

        S é o número de submercados.

        N é a máxima ordem do modelo PAR(p) ajustado.

        P é o número de patamares de carga.

        L é o lag máximo de despacho GNL antecipado.

        :return: Os coeficientes dos cortes em uma tabela.
        :rtype: pd.DataFrame | None
        """
        dados = [r for r in self.data.of_type(SecaoDadosCortes)]
        if len(dados) == 1:
            return dados[0].data
        else:
            return None



MockRegistroPEECadastro = "PEE-CAD  ; 1         ; NEInterior\n"

MockRegistroPEEPotenciaInstaladaPeriodo = (
    "PEE-POT-INST-PER  ;  1        ; 2021/01; 2030/12; 6058.890\n"
)

MockEolicaCadastro = [
    "& PEE-CAD ; Codigo ; Nome ; Identificador ; QuantidadeIConjunto\n",
    "PEE-CAD ; 1 ; NEInterior ;  ; 1\n",
    "PEE-CAD ; 2 ; NELitoral ;  ;  1\n",
    "PEE-CAD ; 3 ; NEPE ;  ;  1\n",
    "PEE-CAD ; 4 ; SULInterior ; ; 1\n",
    "PEE-CAD ; 5 ; SULLitoral ; ; 1\n",
    "\n",
]


MockRegistroPEEConfiguracaoPeriodo = (
    "PEE-CONFIG-PER  ; 1         ; 2021/01; 2030/12; centralizado\n"
)

MockEolicaConfig = [
    "&PEE-CONFIG-PER; CodigoEolica; PeriodoInicial; PeriodoFinal  ; EstadoEolica (centralizado = 1, fixo = 2, nãoexistente = 3)\n",
    "PEE-CONFIG-PER ; 1 ; 2021/01; 2030/12; centralizado\n",
    "PEE-CONFIG-PER ; 2 ; 2021/01; 2030/12; centralizado\n",
    "PEE-CONFIG-PER ; 3 ; 2021/01; 2030/12; centralizado\n",
    "PEE-CONFIG-PER ; 4 ; 2021/01; 2030/12; centralizado\n",
    "PEE-CONFIG-PER ; 5 ; 2021/01; 2030/12; centralizado\n",
]

MockRegistroPEEFTE = "PEE-FPVP-LIN-PU-PER	  ; 1         ; 2021/01; 2030/12; -0.14454132687670500 ; 0.10904637648150100\n"

MockEolicaFTE = [
    "& PEE-FPVP-LIN-PU-PER ; CodEolica;PerIni ;PerFin ; CoeficienteLinear; CoeficienteAngular\n",
    "\n",
    "PEE-FPVP-LIN-PU-PER ; 1; 2021/01; 2030/12;  -0.14454132; 0.10904637\n",
    "PEE-FPVP-LIN-PU-PER ; 2; 2021/01; 2030/12;  -0.34282825; 0.09599283\n",
    "PEE-FPVP-LIN-PU-PER ; 3; 2021/01; 2030/12;  -0.31163518; 0.14003760\n",
    "PEE-FPVP-LIN-PU-PER ; 4; 2021/01; 2030/12;  -0.08251197; 0.07183436\n",
    "PEE-FPVP-LIN-PU-PER ; 5; 2021/01; 2030/12;  -0.24408456; 0.09996426",
]


MockRegistroPEEGeracaoPatamar = (
    "PEE-GER-PROF-PER-PAT  ; 1         ; 2021/01; 2021/01; 1   ; 0.8800\n"
)

MockEolicaGeracao = [
    "& PEE-GER-PROF-PER-PAT; CodEolica;PerIni ;PerFin ;Pat   ;PROFUNDIDADE\n",
    "PEE-GER-PROF-PER-PAT; 1 ; 2021/01; 2021/01; 1 ; 0.8800\n",
    "PEE-GER-PROF-PER-PAT; 1 ; 2021/01; 2021/01; 2 ; 1.0496\n",
    "PEE-GER-PROF-PER-PAT; 1 ; 2021/01; 2021/01; 3 ; 1.0421\n",
    "PEE-GER-PROF-PER-PAT; 1 ; 2021/02; 2021/02; 1 ; 0.8800\n",
    "PEE-GER-PROF-PER-PAT; 1 ; 2021/02; 2021/02; 2 ; 1.0496\n",
    "PEE-GER-PROF-PER-PAT; 1 ; 2021/02; 2021/02; 3 ; 1.0421\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t \n",
]

MockRegistroHistoricoVentoHorizonte = (
    "VENTO-HIST-HORIZ  ; 1979/01     ; 2016/01\n"
)

MockRegistroHistoricoVento = "VENTO-HIST  ; 1           ; 1979/01     ; 1979/02   ;      4.05       ; 1.0\n"


MockRegistroPostoCadastro = ("POSTO-VENTO-CAD  ; 1           ; NEInterior\n",)

MockRegistroPEEPosto = "PEE-POSTO  ; 1         ; 1\n"

MockEolicaPosto = [
    "&POSTO-VENTO-CAD ; CodigoPosto ; NomePosto\n",
    "POSTO-VENTO-CAD  ; 1           ; NEInterior\n",
    "POSTO-VENTO-CAD  ; 2           ; NELitoral\n",
    "POSTO-VENTO-CAD  ; 3           ; NEPE\n",
    "POSTO-VENTO-CAD  ; 4           ; SULInterior\n",
    "POSTO-VENTO-CAD  ; 5           ; SULLitoral\n",
    "\n",
    "&PEE-POSTO ; CodigoPEE ; CodigoPosto\n",
    "PEE-POSTO  ; 1         ; 1\n",
    "PEE-POSTO  ; 2         ; 2\n",
    "PEE-POSTO  ; 3         ; 3\n",
    "PEE-POSTO  ; 4         ; 4\n",
    "PEE-POSTO  ; 5         ; 5\n",
]


MockRegistroPEESubmercado = "PEE-SUBM   ; 1         ; 3\n"

MockPEESubmercado = [
    "& PEE-SUBM; CodigoEolica; CodigoSubmercado\n",
    "PEE-SUBM ; 1 ; 3\n",
    "PEE-SUBM ; 2 ; 3\n",
    "PEE-SUBM ; 3 ; 3\n",
    "PEE-SUBM ; 4 ; 2\n",
    "PEE-SUBM ; 5 ; 2\n",
    "",
]

MockEolica = (
    MockEolicaCadastro
    + MockEolicaConfig
    + MockEolicaFTE
    + MockEolicaGeracao
    + MockEolicaPosto
    + MockPEESubmercado
)



# Imports de módulos externos
from cfinterface.components.section import Section
from typing import IO
import pandas as pd  # type: ignore


class TabelaMediasusit(Section):
    """
    Bloco com o conteúdo da tabela existente no arquivo `MEDIAS-USIT.CSV`.
    """

    __slots__ = ["data"]

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, TabelaMediasusit):
            return False
        bloco: TabelaMediasusit = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO, *args, **kwargs):
        tabela = pd.read_csv(file, skipinitialspace=True)
        col_usi = "USIT_ext"
        tabela = tabela.rename(columns={col_usi: "codigo_usina"})
        tabela = tabela.loc[tabela["codigo_usina"] > 0]
        cols = tabela.columns.tolist()
        df = tabela.drop(columns=[cols[-1]])
        cols_id = ["VAR", "codigo_usina"]
        cols_estagios = [c for c in df.columns.tolist() if c not in cols_id]
        df = df.melt(
            id_vars=cols_id,
            value_vars=cols_estagios,
            var_name="estagio",
            value_name="valor",
        )
        df["estagio"] = df["estagio"].astype(int)
        df["estagio"] -= df["estagio"].min() - 1
        df = df.pivot_table(
            index=["estagio", "codigo_usina"],
            columns="VAR",
            values="valor",
        ).reset_index()
        self.data = df



from inewave.nwlistop.modelos.blocos.ree import REE
from inewave.nwlistop.modelos.arquivos.arquivoreepatamar import (
    ArquivoREEPatamar,
)
from inewave.nwlistop.modelos.viol_ghmin import ViolGhminAnos


class ViolGhmin(ArquivoREEPatamar):
    """
    Armazena os dados das saídas referentes à violação da meta de geração
    hidraulica mínima por patamar, por REE.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `viol_ghmin00x.out`, onde x varia conforme o
    REE em questão.

    """

    BLOCKS = [
        REE,
        ViolGhminAnos,
    ]



import numpy as np  # type: ignore
import pandas as pd  # type: ignore
from datetime import datetime
from inewave.config import MESES_DF
from typing import List

__COLS_IDENTIFICACAO = ["data", "ano", "serie", "patamar", "classe"]


def __formata_df_meses_para_datas_serie(df: pd.DataFrame) -> pd.DataFrame:
    anos = np.array(df["ano"].unique().tolist())
    series = np.array(df["serie"].unique().tolist())
    n_series = len(series)
    datas_df = []
    series_df = []
    for a in anos:
        datas_df.append(
            np.tile(
                pd.date_range(
                    datetime(year=a, month=1, day=1),
                    datetime(year=a, month=12, day=1),
                    freq="MS",
                ).to_numpy(),
                n_series,
            ),
        )
        series_df.append(np.repeat(series, 12))
    valores = []
    for a in anos:
        valores.append(
            df.loc[
                (df["ano"] == a),
                MESES_DF,
            ]
            .to_numpy()
            .flatten()
        )
    df_formatado = pd.DataFrame(
        data={
            "data": np.concatenate(datas_df),
            "serie": np.concatenate(series_df),
            "valor": np.concatenate(valores),
        }
    )
    return df_formatado


def __formata_df_meses_para_datas_serie_patamar(
    df: pd.DataFrame,
) -> pd.DataFrame:
    anos = np.array(df["ano"].unique().tolist())
    series = np.array(df["serie"].unique().tolist())
    patamares = np.array(df["patamar"].unique().tolist())
    n_series = len(series)
    n_patamares = len(patamares)
    datas_df = []
    patamares_df = []
    series_df = []
    for a in anos:
        datas_df.append(
            np.tile(
                pd.date_range(
                    datetime(year=a, month=1, day=1),
                    datetime(year=a, month=12, day=1),
                    freq="MS",
                ).to_numpy(),
                n_series * n_patamares,
            ),
        )
        patamares_df.append(np.tile(np.repeat(patamares, 12), n_series))
        series_df.append(np.repeat(series, 12 * n_patamares))

    valores = []
    for a in anos:
        valores.append(
            df.loc[
                (df["ano"] == a),
                MESES_DF,
            ]
            .to_numpy()
            .flatten()
        )
    df_formatado = pd.DataFrame(
        data={
            "data": np.concatenate(datas_df),
            "patamar": np.concatenate(patamares_df),
            "serie": np.concatenate(series_df),
            "valor": np.concatenate(valores),
        }
    )
    return df_formatado


def __formata_df_meses_para_datas_classetermica_serie_patamar(
    df: pd.DataFrame,
) -> pd.DataFrame:
    anos = np.array(df["ano"].unique().tolist())
    classes = np.array(df["classe"].unique().tolist())
    series = np.array(df["serie"].unique().tolist())
    patamares = np.array(df["patamar"].unique().tolist())
    n_classes = len(classes)
    n_series = len(series)
    n_patamares = len(patamares)
    datas_df = []
    classes_df = []
    series_df = []
    patamares_df = []
    for a in anos:
        datas_df.append(
            np.tile(
                pd.date_range(
                    datetime(year=a, month=1, day=1),
                    datetime(year=a, month=12, day=1),
                    freq="MS",
                ).to_numpy(),
                n_classes * n_series * n_patamares,
            ),
        )
        classes_df.append(np.repeat(classes, 12 * n_series * n_patamares))
        patamares_df.append(
            np.tile(np.repeat(patamares, 12), n_classes * n_series)
        )
        series_df.append(
            np.tile(np.repeat(series, 12 * n_patamares), n_classes)
        )
    valores = []
    for a in anos:
        valores.append(
            df.loc[
                (df["ano"] == a),
                MESES_DF,
            ]
            .to_numpy()
            .flatten()
        )
    df_formatado = pd.DataFrame(
        data={
            "classe": np.concatenate(classes_df),
            "data": np.concatenate(datas_df),
            "patamar": np.concatenate(patamares_df),
            "serie": np.concatenate(series_df),
            "valor": np.concatenate(valores),
        }
    )
    return df_formatado[["classe", "data", "patamar", "serie", "valor"]]


def formata_df_meses_para_datas_nwlistop(df: pd.DataFrame) -> pd.DataFrame:
    colunas_df = df.columns.tolist()
    colunas_identificacao = tuple(
        [c for c in colunas_df if c in __COLS_IDENTIFICACAO]
    )
    mapa_formatacao = {
        ("ano", "serie"): __formata_df_meses_para_datas_serie,
        (
            "ano",
            "serie",
            "patamar",
        ): __formata_df_meses_para_datas_serie_patamar,
        (
            "ano",
            "classe",
            "serie",
            "patamar",
        ): __formata_df_meses_para_datas_classetermica_serie_patamar,
    }
    return mapa_formatacao[colunas_identificacao](df)


def prepara_vetor_anos_tabela(anos: List[str]) -> List[datetime]:
    # Se tem pré, substitui por 0001
    # Se tem pós, substitui por 9999
    # Repete os valores existentes 12 vezes
    anos_convertidos: List[int] = []
    for a in anos:
        if a == "PRE":
            a_convertido = 1
        elif a == "POS":
            a_convertido = 9999
        else:
            a_convertido = int(a)
        anos_convertidos.append(a_convertido)

    anos_array = np.array(anos_convertidos).repeat(len(MESES_DF))
    meses = np.tile(np.arange(1, 13), len(anos))
    return [
        datetime(year=a, month=m, day=1) for a, m in zip(anos_array, meses)
    ]


def repete_vetor(
    valores: list, n_repeticoes: int = len(MESES_DF)
) -> np.ndarray:
    return np.array(valores).repeat(n_repeticoes)


def prepara_valor_ano(ano: int) -> str:
    if ano == 1:
        return "PRE"
    elif ano == 9999:
        return "POS"
    else:
        return str(ano)
